-----------------------------------------------------------------------------
-- Frequency counter
-- *****************
-- Uses a continuous counter with pre scaler
--
-- Inputs:
-- CLK: Main CLK signal at 2^26 Hz
-- PPS: One pulse per second at the GPS 1 second mark
-- COUNT: 26 bit GPS sub second counter
-- RF: RF signal to be counted
--
-- Outputs:
-- FREQ: Frequency output
-- VALID: One cycle valid pulse to indicate a new frequency output
-- UP: Counter value is at or above 16384 Hz
-- FREQA: Direct count without prescaler (diagnostics)
-- FREQB: Direct count from 1:2 prescaler (diagnostics)
-- FREQC: Direct count from 1:4 prescaler (diagnostics)
--
--
-- The RF signal is fed to 3 different prescalers: 1:1, 1:2 and 1:4.
-- Each prescaled RF signal is independently counted using a 32 bit counter
-- Each of these counters is latched ~15 us prior to the 1 second mark
-- 4 cycles prior to the 1 second mark each counter is read out,
-- and the difference to the previous value is formed.
-- The difference is the new frequency count muliplied by the prescaler factor
-- If the frequency as measured by the 1:4 prescaled signal is higher than
-- 2^28 Hz (268 MHz) it is used.
-- If lower, but higher than 2^27 Hz (134 MHz), the 2:1 prescaled frequency
-- count is used.
-- Below 2^27 Hz (134 MHz), but above 2^19 Hz (0.5 MHz) the direct frequency
-- count is used.
-- Below 2^19 Hz (0.5 MHz) the count will be set to zero.
-- The valid bit goes high at the 1 second mark for one cycle
-- The up bit goes high for an entire second, if the frequency is above
-- 0.5 MHz.
--
-- None of the prescalers or RF counters is ever reset. This means the
-- frequency accuracy will scale linearly with the measurement time and
-- is one, two and four cycles for frequency below 134 MHz, for
-- frequencies between 134 MHz and 268 MHz and for frequencies above 268 MHz,
-- respectively.
--
-- Example: A frequency of 300 MHz determined by averaging over 8 seconds
--          will have an error of 0.5 Hz.
--
-- Daniel Sigg
-- 17 December, 2014
-----------------------------------------------------------------------------

library IEEE;
use IEEE.Std_Logic_1164.all;
use IEEE.Numeric_Std.all;

entity FrequencyCounter is port
    (
    CLK : in std_logic;
    PPS : in std_logic;
    COUNT : in std_logic_vector (25 downto 0);
    RF : in std_logic;
    FREQ : out std_logic_vector (31 downto 0);
    VALID : out std_logic;
    UP: out std_logic;
    FREQA : out std_logic_vector (31 downto 0);
    FREQB : out std_logic_vector (31 downto 0);
    FREQC : out std_logic_vector (31 downto 0)
    );
end entity FrequencyCounter;

architecture ark of FrequencyCounter is

     type counter_bank is array (2 downto 0) of unsigned (27 downto 0);
     type trigger_bank is array (2 downto 0) of std_logic_vector (2 downto 0);

     signal oldpps : std_logic := '0';
     signal precount : unsigned (25 downto 0) := (others => '0');
     signal trig : std_logic;
     signal readenable : std_logic;

     signal PreScaler2 : std_logic := '0';
     signal PreScaler4 : std_logic := '0';
     signal pre : std_logic_vector (2 downto 0) := B"000";

     signal counters : counter_bank := (others => (others => '0'));
     signal pretrig : trigger_bank := (others => B"000");
     signal trigcounts : counter_bank := (others => (others => '0'));
     signal curfreq : counter_bank := (others => (others => '0'));
     signal oldfreq : counter_bank := (others => (others => '0'));
     signal newfreq : counter_bank := (others => (others => '0'));

begin
     -- advanced count in order to trigger 1024 cycles ahead of 1 PPS
     precount <= unsigned (COUNT) + to_unsigned (1024, 26);
     precounter : process (CLK, precount) is
     begin
          if rising_edge (CLK) then
             trig <= not precount (25);
          end if;
     end process precounter;

     -- readout enable
     readenable <= '1' when (signed (COUNT) = to_signed (-4, 26)) else '0';

     -- no prescaler
     pre(0) <= RF;

     -- 2:1 prescaler
     prescale2 : process (RF, PreScaler2) is
     begin
          if rising_edge (RF) then
             PreScaler2 <= not PreScaler2;
          end if;
          pre(1) <= not PreScaler2;
     end process prescale2;

     -- 4:1 prescaler (2 Johnson counters in series)
     prescale4 : process (pre, PreScaler4) is
     begin
          if rising_edge (pre(1)) then
             PreScaler4 <= not PreScaler4;
          end if;
          pre(2) <= not PreScaler4;
     end process prescale4;

     -- counter bank (one counter for each prescaler)
     generate_counters: for i in 0 to 2 generate
     begin
          -- count the frequency of the RF signal after the prescaler.
          count : process (pre, counters) is
          begin
               if rising_edge (pre(i)) then
                    counters(i) <= counters(i) + 1;
               end if;
          end process count;

          -- Synchronize trigger with prescaler
          synctrig : process (pre, trig, pretrig) is
          begin
               if rising_edge (pre(i)) then
                  pretrig(i)(0) <= trig;
                  pretrig(i)(1) <= pretrig(i)(0);
               end if;
               pretrig(i)(2) <= pretrig(i)(0) and not pretrig(i)(1);
          end process synctrig;

          -- latch counters every 1 second using the (pre) trigger
          synccount : process (pre, pretrig, counters) is
          begin
               if rising_edge (pre(i)) then
                  if (pretrig(i)(2) = '1') then
                     trigcounts(i) <= counters(i);
                  end if;
               end if;
          end process synccount;

          -- read counter on enable (we are back in the CLK domain)
          readcount : process (CLK, readenable, trigcounts) is
          begin
               if rising_edge (CLK) then
                  if (readenable = '1') then
                     curfreq(i) <= trigcounts(i);
                     oldfreq(i) <= curfreq(i);
                  end if;
               end if;
          end process readcount;
          newfreq(i) <= curfreq(i) - oldfreq(i);

     end generate generate_counters;

     selectfreq : process (newfreq) is
     begin
          -- use prescaled-by-4 counter, if frequency is larger than 268 MHz
          if (newfreq(2) >= to_unsigned (1, 32) sll 26) then
             -- multiply by 4
             FREQ <=  B"00" & std_logic_vector (newfreq(2)) & B"00";
             UP <= '1';
          -- use prescaled-by-2 counter, if frequency is larger than 134 MHz
          elsif (newfreq(2) >= to_unsigned (1, 32) sll 25) then
             -- multiply by 2
             FREQ <=  B"000" & std_logic_vector (newfreq(1)) & B"0";
             UP <= '1';
          -- use direct counter, if frequency is larger than 0.5 MHz
          elsif (newfreq(2) >= to_unsigned (1, 32) sll 17) then
             -- multiply by 1
             FREQ <=  B"0000" & std_logic_vector (newfreq(0));
             UP <= '1';
          -- set to zero, if below 0.5 MHz (RF input is AC coupled)
          else
             FREQ <= (others => '0');
             UP <= '0';
          end if;
     end process selectfreq;

     -- set output values
     FREQA <= B"0000" & std_logic_vector (newfreq(0));
     FREQB <= B"000" & std_logic_vector (newfreq(1)) & B"0";
     FREQC <= B"00" & std_logic_vector (newfreq(2)) & B"00";

     oldpps <= PPS when rising_edge (CLK);
     VALID <= '1' when ((PPS = '1') and (oldpps = '0')) else '0';

end architecture ark;
